/*
 * fillin_settings()
 * is_dangerous_settings()
 * send_arp_packet()
 *
 * Copyright (c) 2001-2002, The Trustees of Princeton University, All Rights Reserved.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "defs.h"
#include "arpsend.h"

/* for libnet */
char libnet_errbuf[LIBNET_ERRBUF_SIZE];
struct libnet_link_int *nd = NULL;

u_char eaddr_broadcast[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};


/* Fill in settings, taking into account args and options.
   Also opens nd, the link interface.
*/
void 
fillin_settings(void)
{

	if (interface_name == NULL) {
		/* ask libnet to select the first interface that is up (other than loopback) */
		struct sockaddr_in sin;
		if (libnet_select_device(&sin, &interface_name, libnet_errbuf) == -1) {
			fprintf(stderr, "%s: error selecting an interface: %s\n", prog, libnet_errbuf);
			cleanup();
			exit(1);
		}
	}

	if ((nd = libnet_open_link_interface(interface_name, libnet_errbuf)) == NULL) {
		fprintf(stderr, "%s: error opening interface '%s': %s\n", prog, interface_name, libnet_errbuf);
		cleanup();
		exit(1);
	}

	if (!sender_protocol_address_set) {
		/* sender_protocol_address defaults to that of outgoing interface */
		/* Note libnet_get_ipaddr() returns host-byte order.  Versions of libnet older than 1.0.2
		   incorrectly document it as returning network-byte order. */
		if ((sender_protocol_address.s_addr = htonl(libnet_get_ipaddr(nd, interface_name, libnet_errbuf))) == 0) {
			fprintf(stderr, "%s: can't determine IP address of interface '%s': %s\n", prog, interface_name, libnet_errbuf);
			cleanup();
			exit(1);
		}
	}

	if (!ether_source_address_set || !sender_hardware_address_set) {
		struct ether_addr *my_eaddr;

		/* set my_eaddr to that of the outgoing interface */
		if ((my_eaddr =libnet_get_hwaddr(nd, interface_name, libnet_errbuf)) == NULL) {
			fprintf(stderr, "%s: can't determine hardware address of interface '%s': %s\n", prog, interface_name, libnet_errbuf);
			cleanup();
			exit(1);
		}

		if (!ether_source_address_set) {
			bcopy(my_eaddr, &ether_source_address, sizeof(ether_source_address));
		}
		if (!sender_hardware_address_set) {
			bcopy(my_eaddr, &sender_hardware_address, sizeof(sender_hardware_address));
		}
	}

	if (!ether_dest_address_set) {
		/* ether_dest_address defaults to Ethernet broadcast address */
		bcopy(&eaddr_broadcast, &ether_dest_address, sizeof(ether_dest_address));
	}

	if (!target_hardware_address_set) {
		/* target_hardware_address defaults to 0 */
		bzero(&target_hardware_address, sizeof(target_hardware_address));
	}

	if (!arp_opcode_set) {
		/* arp_opcode defaults to ARPOP_REQUEST */
		arp_opcode = ARPOP_REQUEST;
	}

	
	if (!quiet) {
		fprintf(stdout, "Using settings:\n");
		fprintf(stdout, "  interface %s\n", interface_name);
		fprintf(stdout, "  ether_source_address %s\n", ether_ntoa(&ether_source_address));
		fprintf(stdout, "  ether_dest_address %s\n", ether_ntoa(&ether_dest_address));
		fprintf(stdout, "  arp_opcode %u (%s)\n", arp_opcode,
				arp_opcode==ARPOP_REQUEST ? "request" : (arp_opcode == ARPOP_REPLY ? "reply" : "unknown"));
		fprintf(stdout, "  sender_hardware_address %s\n", ether_ntoa(&sender_hardware_address));
		fprintf(stdout, "  sender_protocol_address %s\n", inet_ntoa(sender_protocol_address));
		fprintf(stdout, "  target_hardware_address %s\n", ether_ntoa(&target_hardware_address));
		fprintf(stdout, "  target_protocol_address %s\n", inet_ntoa(target_protocol_address));
	}
	
	return;
}


/* Check for dangerous settings; return 1 if any found, else return 0.
   Print a warning for each dangerous item found.
 */
int
is_dangerous_settings(void)
{
	int danger_found = 0;

	/* An ether_src should never be multicast or broadcast.
	   Imagine what happens if bridges/switches learn ff:ff:ff:ff:ff:ff is located on a specific port!
	*/
	if (ether_source_address.ether_addr_octet[0] & 0x01) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: ether_source_address '%s' is broadcast or multicast\n(not a valid Ethernet source).\n", 
			ether_ntoa(&ether_source_address));
		fprintf(stdout, "A careless switch/bridge might mistakenly insert this into its\nforwarding table.\n");
		danger_found++;
	}

	/* sender_hardware_address shouldn't be multicast.  
	   Regardless of arp_opcode, listeners may glean (spa,sha) into their arp caches.
	   And if arp_opcode is request, clients may send respond packets to sender_hardware_address.
	*/
	if (sender_hardware_address.ether_addr_octet[0] & 0x01) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: sender_hardware_address '%s' is broadcast or multicast.\n", 
			ether_ntoa(&sender_hardware_address));
		fprintf(stdout, "Recipients of the frame (ether_dest_address=%s) that glean the\n(sender_hardware_address,sender_protocol_address) tuple may be left with a bad\nARP cache entry for the sender_protocol_address (%s).\n", ether_ntoa(&ether_dest_address), inet_ntoa(sender_protocol_address));
		if (arp_opcode == ARPOP_REQUEST) {
			fprintf(stdout, "Also, any recipients that respond to the ARP Request will send their responses\nto this Ethernet multicast/broadcast address.\n");
		}
		danger_found++;
	}

	if (IN_CLASSD(target_protocol_address.s_addr)) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: target_protocol_address '%s' is an IP multicast address.\n", inet_ntoa(target_protocol_address));
		danger_found++;
	}
	if (target_protocol_address.s_addr == INADDR_BROADCAST) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: target_protocol_address '%s' is the IP limited\nbroadcast address.\n", inet_ntoa(target_protocol_address));
		danger_found++;
	}
	if (IN_EXPERIMENTAL(target_protocol_address.s_addr)) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: target_protocol_address '%s' is an IP experimental (Class E) address.\n", inet_ntoa(target_protocol_address));
		danger_found++;
	}
	if (MY_IN_LOOPBACK(target_protocol_address.s_addr)) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: target_protocol_address '%s' is on the loopback network.\n", inet_ntoa(target_protocol_address));
		danger_found++;
	}
	/* XXX we should also check to see if target_protocol_address is any of the other flavors of IP broadcast
	   addresses valid on this network.  
	*/

	if (IN_CLASSD(sender_protocol_address.s_addr)) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: sender_protocol_address '%s' is an IP multicast address.\n", inet_ntoa(sender_protocol_address));
		fprintf(stdout, "Recipients of the frame (ether_dest_address=%s) that glean the\n(sender_hardware_address,sender_protocol_address) tuple may be left with a bad\nARP cache entry for the sender_protocol_address.\n", ether_ntoa(&ether_dest_address));
		danger_found++;
	}
	if (sender_protocol_address.s_addr == INADDR_BROADCAST) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: sender_protocol_address '%s' is the IP limited\nbroadcast address.\n", inet_ntoa(sender_protocol_address));
		fprintf(stdout, "Recipients of the frame (ether_dest_address=%s) that glean the\n(sender_hardware_address, sender_protocol_address) tuple may be left with a bad\nARP cache entry for the IP limited broadcast address.\n", ether_ntoa(&ether_dest_address));
		danger_found++;
	}
	if (IN_EXPERIMENTAL(sender_protocol_address.s_addr)) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: sender_protocol_address '%s' is an IP experimental (Class E) address.\n", inet_ntoa(sender_protocol_address));
		fprintf(stdout, "Recipients of the frame (ether_dest_address=%s) that glean the\n(sender_hardware_address,sender_protocol_address) tuple may be left with a bad\nARP cache entry for the sender_protocol_address.\n", ether_ntoa(&ether_dest_address));
		danger_found++;
	}
	if (MY_IN_LOOPBACK(sender_protocol_address.s_addr)) {
		if (danger_found) /* this is not our first warning */
			fprintf(stdout, "\n");
		fprintf(stdout, "*** Warning: sender_protocol_address '%s' is on the loopback network.\n", inet_ntoa(sender_protocol_address));
		danger_found++;
	}
	/* XXX we should also check to see if sender_protocol_address is any of the other flavors of IP broadcast
	   addresses valid on this network.  
	*/
	

	{
		struct in_addr my_ipaddr;
		struct ether_addr *my_eaddr; /* XXX libnet_get_hwaddr() returns ptr to static storage */

		if (((my_eaddr = libnet_get_hwaddr(nd, interface_name, libnet_errbuf)) != NULL) &&
			(((my_ipaddr.s_addr = libnet_get_ipaddr(nd, interface_name, libnet_errbuf))) != 0)) {

			/* we were able to determine my_eaddr and my_eaddr for this interface */
	
			if ((sender_protocol_address.s_addr == my_ipaddr.s_addr) && bcmp(&sender_hardware_address, my_eaddr, sizeof(my_eaddr))) {
				if (danger_found) /* this is not our first warning */
					fprintf(stdout, "\n");
				fprintf(stdout, "*** Warning: sender_protocol_address '%s' is my IP address, but\nsender_hardware_address '%s' ",
						inet_ntoa(sender_protocol_address), ether_ntoa(&sender_hardware_address));
				/* ether_ntoa() returns static storage, break calls into separate statements */
				fprintf(stdout, "is not my Ethernet address\n('%s').\n", ether_ntoa(my_eaddr));
				fprintf(stdout, "Recipients of the frame (ether_dest_address=%s) that glean the\n(sender_hardware_address,sender_protocol_address) tuple may be left with a bad\nARP cache entry for my IP address.\n", ether_ntoa(&ether_dest_address));
				danger_found++;
			}
		}
	}
	
	return danger_found;
}


/* construct and send ARP packet.
   Return 0 on success, 1 on failure.
*/
int
send_arp_packet(void)
{
	u_char *write_packet = NULL;								/* packet buffer for writing */
	int ether_arp_packet_size = LIBNET_ETH_H + LIBNET_ARP_H;	/* size of packet we will write */
	int bytes_written;


	/* allocate packet buffer space and zero it */
	if (libnet_init_packet(ether_arp_packet_size, &write_packet) == -1) {
		fprintf(stderr, "%s: libnet_init_packet failed\n", prog);
		if (write_packet != NULL) /* just in case something got allocated */
			libnet_destroy_packet(&write_packet);
		return(0); /* failure */
	}

	/* build the Ethernet header */
	if (libnet_build_ethernet(
		ether_dest_address.ether_addr_octet,		/* ether_dst */
		ether_source_address.ether_addr_octet,		/* ether_src */
		ETHERTYPE_ARP,								/* ethertype */
		NULL, 0,									/* no optional payload */
		write_packet								/* start of ethernet header in packet */
	) < 0) {
		fprintf(stderr, "%s: libnet_build_ethernet failed\n", prog);
		libnet_destroy_packet(&write_packet);
		return(0);
	}

	/* build the ARP header and payload */
	if (libnet_build_arp(
		ARPHRD_ETHER,								/* hardware type */
		ETHERTYPE_IP,								/* proto type */
		6,											/* haddr len, 6 == Ethernet */
		4,											/* proto len, 4 == IPv4 */
		arp_opcode,									/* arp_opcode == ARPOP_REQUEST or ARPOP_REPLY */
		sender_hardware_address.ether_addr_octet,	/* arp_sha */
		(u_char *) &sender_protocol_address.s_addr,	/* arp_spa */
		target_hardware_address.ether_addr_octet,	/* arp_tha */
		(u_char *) &target_protocol_address.s_addr,	/* arp_tpa */
		NULL, 0,									/* no optional payload */
		write_packet + LIBNET_ETH_H					/* start of arp header in packet */
	) < 0) {
		fprintf(stderr, "%s: libnet_build_arp failed\n", prog);
		libnet_destroy_packet(&write_packet);
		return(0);
	}

	/* write the packet */
	bytes_written =  libnet_write_link_layer(nd, interface_name, write_packet, ether_arp_packet_size);
	if (bytes_written < ether_arp_packet_size)
		fprintf(stderr, "%s: libnet_write_link_layer: bytes written: %d (expected %d)\n", prog, bytes_written, ether_arp_packet_size);

	libnet_destroy_packet(&write_packet);

	return(1); /* success */
}
